# Standard Libraries
from collections.abc import Iterable
from typing import Iterable

# Third Party Libraries
import psycopg2
from psycopg2.extras import NamedTupleCursor

from .api import thesoapi
from .exceptions import ThesorimedError


class ThesoItem:
    """
    Base classe interface pour thesorimed
    """

    def __init__(self, host, dbname, user, password):
        self.host = host
        self.dbname = dbname
        self.user = user
        self.password = password

    def _connect(self):
        """
        Base fonction to connect to database.
        Return a pscipg connection
        should return psycopg2.connect
        """
        # raise NotImplementedError
        return psycopg2.connect(
            host=self.host, dbname=self.dbname, user=self.user, password=self.password)

    def _normalize_req(self, obj_api, req):
        req = list(req)
        if isinstance(req[0], str):
            return req
        if obj_api.genre == "char":
            req[0] = str(req[0])
        elif obj_api.genre == "cursor":
            req[0] = ','.join(map(str, req[0]))
        else:
            raise ThesorimedError('genre invalide')
        return req

    def _appel_proc(self, obj_api, req):

        if 'str' in obj_api.input_type[0]:
            requete = self._normalize_req(obj_api, req)
        else:
            requete = req

        with self._connect() as con:
            with con.cursor(cursor_factory=NamedTupleCursor) as curs:
                curs.execute("SET search_path TO thesorimed, public")
                curs.callproc("thesorimed." + obj_api.name, requete)
                res = curs.fetchone()

                try:
                    result = getattr(res, obj_api.name).split(', ')
                except AttributeError:
                    return None

                if result[0].startswith("<unnamed portal"):
                    f = f'FETCH ALL IN "{result[0]}"'
                    curs.execute(f)
                    result = curs.fetchall()

                return result

    @staticmethod
    def _valide_req(obj, req):
        if len(req) != len(obj.input_type):
            raise ThesorimedError("Le nombre d'argument est invalide")

        if obj.input_type == ['str', 'int']:
            if not isinstance(req[0], Iterable):
                print(req)
                raise ThesorimedError("Le premier argument doit être iterable")
            for item in req[0]:
                try:
                    int(item)
                except ValueError:
                    raise ThesorimedError("L'argument attendu est une liste d'entier")

        for x, y in zip(obj.input_type, req):
            if x.startswith('int'):
                if type(y) != int:
                    raise ThesorimedError("L'argument doit être un entier")
            import re
            longueur_champs = re.findall(r"(?:int|str)([0-9]+)", x)
            if longueur_champs:
                if y > pow(10, int(longueur_champs[0])):
                    raise ThesorimedError(f"Longueur de requête limité à {longueur_champs[0]}")
        return True

    def proc(self, name, *req):
        try:
            obj_api = thesoapi[name]
        except KeyError:
            raise ThesorimedError("La procédure appelée n'existe pas")

        self._valide_req(obj_api, req)

        return self._appel_proc(obj_api, req)


"""
usage dans Django:

from theso import ThesoItem

class Medicament:
    name
    cip

    detail(self, name, req):
        a = ThesoItem(self.cip)
        return a.proc(name, req)

    def monographie:
        retruen a.ThesoItem(self.cip).monographe


"""



# Third Party Libraries
import pytest

# pythonthesorimed
from pythonthesorimed.api import ProcApi, thesoapi
from pythonthesorimed.exceptions import ThesorimedError
from pythonthesorimed.thesoitem import ThesoItem

instance = ThesoItem("1", "3", "4", "5")


class TestValidReq:
    def test_has_not_equals_args_number_raises(self):
        with pytest.raises(ThesorimedError) as e:
            ThesoItem._valide_req(thesoapi['get_cip'], [1, 2, 3])
        assert str(e.value) == "Le nombre d'argument est invalide"

    def test_has_equals_argq_number_pass(self):
        assert ThesoItem._valide_req(thesoapi['get_cip'], [1])

    def test_str_int_is_iterable(self):
        assert ThesoItem._valide_req(thesoapi['get_the_spe_details'],
                                     [(1, 2), 1])

    def test_str_int_not_iterable_via_int(self):
        with pytest.raises(ThesorimedError) as e:
            ThesoItem._valide_req(thesoapi['get_the_spe_details'], [1, 1])
        assert str(e.value) == "Le premier argument doit être iterable"

    def test_str_int_is_integer_list(self):
        assert ThesoItem._valide_req(thesoapi['get_the_spe_details'],
                                     [(1, 2), 1])

    def test_str_int_not_list_of_integer(self):
        with pytest.raises(ThesorimedError) as e:
            assert ThesoItem._valide_req(thesoapi['get_the_spe_details'],
                                         [("o", "oi"), 1])
        assert str(e.value) == "L'argument attendu est une liste d'entier"

    def test_int_alone_is_not_integer_raises(self):
        with pytest.raises(ThesorimedError) as e:
            assert ThesoItem._valide_req(thesoapi['get_the_cim10'], ["lij"])
        assert str(e.value) == "L'argument doit être un entier"

    # def test_int_alone_with_lentgh_is_not_integer_raises(self):
    #     with pytest.raises(ThesorimedError) as e:
    #         assert ThesoItem._valide_req(thesoapi['get_cip'], ["lij"])
    #     assert str(e.value) == "L'argument doit être un entier"

    def test_longueur_req_trop_grande_raises(self):
        with pytest.raises(ThesorimedError) as e:
            assert ThesoItem._valide_req(thesoapi['get_cip'], [1234567])
        assert str(e.value) == "Longueur de requête limité à 6"

    def test_longueur_ok_pass(self):
        assert ThesoItem._valide_req(thesoapi['get_cip'], [123456])


@pytest.mark.parametrize(
    'name,req,expected',
    [('get_cons', [1], ['1']), ('get_the_spe_details', [[1], 1], ['1', 1]),
     ('get_the_info_spe', [[3, 4, 5, 6, 7, 8], 1], ["3,4,5,6,7,8", 1]),
     ('get_the_code_cim10', ["bla"], ["bla"])])
def test_normalize_pass(name, req, expected):
    assert instance._normalize_req(thesoapi[name], req) == expected


def test_normalize_raises():
    a = list(thesoapi['get_cip'])
    a[2] = "fake"
    nt = ProcApi(*a)

    with pytest.raises(ThesorimedError) as e:
        instance._normalize_req(nt, [1])
    assert str(e.value) == "genre invalide"


def test_proc_raises():
    with pytest.raises(ThesorimedError) as e:
        instance.proc("blabalbal", [1])
    assert str(e.value) == "La procédure appelée n'existe pas"



# Third Party Libraries
import pytest

# pythonthesorimed
from pythonthesorimed.exceptions import ThesorimedError
from pythonthesorimed.thesoitem import ThesoItem

base_get_cip = {
    1: ['3400933354978'],
    2: None,
    3: ['3400930916292'],
    8: ['3400933449582', '3400936003736', '3400937224321'],
    9: None
}

base_is_atu = {
    7087: 'O',
    8849: 'N',
}

instance = ThesoItem("localhost", "thesorimed", "j", "j")

# test appel_cher


def test_get_cip():
    for i in [1, 2, 3, 8, 9]:
        assert instance.proc('get_cip', i) == base_get_cip[i]


def test_is_atu():
    for k, v in base_is_atu.items():
        assert instance.proc('is_atu', k) == [v]


# test _refcursor
def test_get_the_code_cim10():
    assert instance.proc('get_the_code_cim10',
                         'INSOMNIE')[0].cim_code == 'G47.0'


def test_get_the_gen_equiv():
    r = instance.proc('get_the_gen_equiv', 3, 1)
    assert [list(i) for i in r] == [[
        'RIFADINE 300MG GELULE', 4793, 'Rifampicine 300 mg gelule',
        'RIMACTAN 300MG GELULE', 3
    ], [
        'RIMACTAN 300MG GELULE', 3, 'Rifampicine 300 mg gelule',
        'RIMACTAN 300MG GELULE', 3
    ]]


def test_proc_launch_validation():
    with pytest.raises(ThesorimedError):
        instance.proc('get_cip', 1, 2, 3)


def test_appel_proc_check_string():
    pass
